/*
** Module   : DmoWriteStorage.java
** Abstract : A storage for the previously saved versions of a DMO.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -------------------------------- Description --------------------------------
** 001 OM  20140430 Created initial implementation with minimum required functionality. 
*/
package com.goldencode.p2j.persist;

import com.goldencode.p2j.util.*;
import java.util.*;

/**
 * This class maintains the history of a DMO old versions saved by persistence. The values of 
 * OLD BUFFER in a WRITE trigger is not mandatory the same with the snapshot of the 
 * <code>RecordBuffer</code>. They are rather unrelated: the snapshot will get the last value 
 * after an UNDO and can contain eventual changes since tha last WRITE, but this class keeps the 
 * stack per DMOs, that means that all referring buffers will see the last saved version of the
 * record. Calling <code>VALIDATE</code> on a buffer will make the DMO to be written to 
 * persistence and all buffers referring that dmo will appear as 'valid' so subsequent calls
 * of <code>VALIDATE</code> on them will have no effect. 
 * 
 * NOTE:
 *    rename the class to Dmo Write Details or something more appropriate 
 */
public class DmoWriteStorage
{
   /**
    * The last persisted record to database. This is needed for WRITE trigger to be optionally 
    * passed as old buffer value and checking if the content of the current record
    * is changed when rollbacking. It may be null if the buffer is not initialized or persisted.
    */
   private Persistable lastPersistedDMO = null;

   /** Flags that this buffer was changed since last write to database. */
   private boolean touched = false;

   /**
    * This stack contains the data needed for flush method in order to decide if the WRITE
    * database events / triggers should be fired and management in the case of an UNDO statement.
    * The lastPersistedValue and wasTouched evolve independently of the current executing block
    * but in the event of UNDO they are restored from this stack.
    */
   private LinkedList<WriteSnapshot> flushSnapshotData = new LinkedList<>();
   
   /**
    * The name of the implementation of the record. It is kept just for logging purpose.
    */
   private String recordName;
   
   /**
    * The constructor for the structure.
    * 
    * @param   className
    *          The class name of the dmo implementation of the record. Only used for logging.
    */
   public DmoWriteStorage(String className)
   {
      this.recordName = className;
      logMe("created", "");
   }
   
   /**
    * Rollbacks the dmo to the previous state before entering the last block. This is an effect 
    * of an undo operation. The <code>lastPersistedDMO</code> and <code>touched</code> are 
    * reverted to their values before the block was started. <code>lastPersistedDMO</code> will 
    * be the previously saved value (usually not the value of the record at the start of the 
    * rollbacked block). <code>touched</code> flags if the record has at least one field changed 
    * since the WRITE event until the block started.
    * 
    * @param   persistence
    *          The persistence.
    * @param   blockDepth
    *          The block depth that is rollbacked.
    *          
    * @return  <code>true</code> is some data is written to persistence and it should be flushed 
    *          to database.
    */
   public boolean rollback(Persistence persistence, int blockDepth)
   {
      logMe("rollback", blockDepth + "");
      
      if (flushSnapshotData.isEmpty())
      {
         logMe("removing ? ", "no such block " + blockDepth);
         
         /* roll-backing to an upper block than [flushSnapshotData] contains means to 
            go to initial state. (Only one such event should happen ?) */
         lastPersistedDMO = null;
         touched = false;
         return false;
      }

      boolean forceFlush = false; 
      final WriteSnapshot flushNode = flushSnapshotData.getLast(); // the state before entering this block
      assert (flushNode.blockDepth == blockDepth); // sanity check

      // NOTE: we are intentionally using the == operator. There is no need to check their
      // content bit-by-bit because the DMO snapshot is built from the scratch each time
      // the flush method decides to save the DMO to persistence
      if (lastPersistedDMO == flushNode.persistedDMO)
      {
         // no commits of the record in this block ? restore the wasTouched from upper block
         touched = flushNode.dirty;
      }
      else
      {
         // there was at least one flush/commit to database, restoring the old content:

         // rolling back to value at the beginning of the block
         lastPersistedDMO = flushNode.persistedDMO;
         touched = flushNode.dirty;

         // if lastPersistedDMO == null then the record was newly created,
         // nothing to restore to database, the ReversibleCreate will handle if the case
         if (lastPersistedDMO != null)
         {
            boolean commit = false;
            try
            {
               if (false) // disable silent db rollbacking
               {
                  // Persist back the old snapshot.
                  commit = persistence.beginTransaction();

                  // TODO: using Session.save() will result in a "duplicate key value violates 
                  // unique constraint"
                  // TODO: using Session.saveOrUpdate() assignment will fail.
                  // this is normal as the lastPersistedDMO was not registered with persistence, 
                  // only the true dmo from which lastPersistedDMO was cloned can be saved   
                  persistence.save(lastPersistedDMO, lastPersistedDMO.getId());
               }

               // validation is not be necessary as the currently saved record was persisted 
               // successfully some time ago
               forceFlush = !true;
            }
            catch (PersistenceException exc)
            {
               if (commit)
               {
                  commit = false;
                  try
                  {
                     persistence.rollback();
                  }
                  catch (PersistenceException ignored)
                  {
                  }
               }
               ErrorManager.recordOrThrowError(exc);
            }
            finally
            {
               if (commit)
               {
                  try
                  {
                     persistence.commit();
                  }
                  catch (PersistenceException exc)
                  {
                     ErrorManager.recordOrThrowError(exc);
                  }
               }
            }
         }
      }
      return forceFlush;
   }

   /**
    * Mark this DMO as being touched. A deep copy of the current state of the DMO is kept. It 
    * will be used for checking if the DMO changes again and also as READ-ONLY old-buffer 
    * in the case the WRITE trigger is fired.
    * 
    * @param   currentRecord
    *          The current dmo state.
    */
   public void touch(Persistable currentRecord)
   {
      setTouched(true);

      if (lastPersistedDMO == null)
      {
         // very lazy initialization, this is the first time the record was changed
         lastPersistedDMO = (Persistable) currentRecord.deepCopy();
      }
      logMe("touch", currentRecord.toString());
   }

   /**
    * Obtain the state of the DMO that was last flushed to database.
    * 
    * @return  The last persisted value of the DMO.
    */
   public Persistable getLastPersistedDMO()
   {
      return lastPersistedDMO;
   }

   /**
    * Checks if the DMO was touched (any of the field of the record was modified) since the last
    * flush to backing database. Setting back all the fields to the last persisted value of the
    * DMO will not invalidate the <code>touched</code> flag.
    * 
    * @return  <code>true</code> if the DMO was altered since it was last saved.
    */
   public boolean wasTouched()
   {
      return touched;
   }

   /**
    * Sets / updates the last version of the DMO. This is only called after a successful save.
    * 
    * @param   lastPersistedDMO
    *          A deep copy of the record just flushed to database.
    */
   public void setLastPersistedDMO(Persistable lastPersistedDMO)
   {
      logMe("setLastPersistedDMO", lastPersistedDMO.toString());
      this.lastPersistedDMO = lastPersistedDMO;
   }

   /**
    * Mark the record as being touched (or not). After the dmo is saved to persistence, it is 
    * marked as clean (<code>touched = false</code>) so an eventual <code>flush</code> will be 
    * ignored until record is cltered again.   
    * 
    * @param   touched
    *          <code>true</code> if the record was altered.
    */
   public void setTouched(boolean touched)
   {
      logMe("setWasTouched", touched + "");
      
      this.touched = touched;
   }
   
   /**
    * Checks if a flush is necessary for this DMO. This is true when two conditions are met 
    * simultaneously: the current DMO was touched and lastPersistedDMO is not null.
    * 
    * @return  <code>true</code> if the DMO should be flushed to database
    */
   public boolean needFlush()
   {
      return (lastPersistedDMO != null) && touched;
   }
   
   /**
    * Notification that a new block is about to start. The current DMO flush-state is pushed on
    * the stack.
    * 
    * @param   blockDepth
    *          The block depth at which the snapshot is taken (usually this is greater than the
    *          internal stack depth).
    */
   public void enterBlock(int blockDepth)
   {
      logMe("enterBlock", blockDepth + "");
      flushSnapshotData.addLast(new WriteSnapshot(blockDepth, touched, lastPersistedDMO));
   }
   
   /**
    * Notification that the currently processed block is finished. The last DMO flush-state is 
    * popped removed the stack.
    *
    * @param   blockDepth
    *          The block depth that ended.
    */
   public void exitBlock(int blockDepth)
   {
      logMe("exitBlock", blockDepth + "");
      if (flushSnapshotData.isEmpty())
      {
         // unregister me, probably out of scope!
         logMe("removing ? ", "no such block " + blockDepth);
         
         return;
      }
      final WriteSnapshot last = flushSnapshotData.removeLast();
      assert (last.blockDepth == blockDepth);
   }
   
   /**
    * Private logging utility function.
    * 
    * @param   method
    *          The method from within the log takes place.
    * @param   info1
    *          Information on the event being processed.
    */
   private void logMe(String method, String info1)
   {
//      System.err.println("DMOStorage::" + method + "(" + recordName + ", " + info1 + "): " +
//             (lastPersistedDMO == null ? "null" : lastPersistedDMO.getId().toString()));
   }
   
   
   /**
    * This immutable class is an element in the stack of undo nodes for previous persisted DMOs.
    * The objects of this kind are used only in a rollback operation in order to obtain the exact
    * configuration of the OLD BUFFER in a WRITE trigger. Because WRITE event is not directly
    * related to transactions block we cannot rely on the snapshot field of RecordBuffer.
    */
   class WriteSnapshot
   {
      /**
       * The block depth (BD) at which the snapshot is taken. This field of the class is just
       * for debugging purposes.
       */
      private int blockDepth;

      /**
       *  Flag for checking if the record touched before entering block (in BD-1).
       */
      private boolean dirty;

      /**
       * The value of lastPersistedValue entering block (in BD-1).
       */
      private Persistable persistedDMO;

      /**
       * The constructor that builds the object.
       *
       * @param   blockDepth
       *          The block depth (BD) at which the snapshot is taken.
       * @param   dirty
       *          Flags if the record touched before entering block.
       * @param   persistedDMO
       *          The value of lastPersistedValue entering block.
       */
      WriteSnapshot(int blockDepth, boolean dirty, Persistable persistedDMO)
      {
         this.blockDepth = blockDepth;
         this.dirty = dirty;
         this.persistedDMO = persistedDMO;
      }
   }
}


